import os
import pandas as pd


def blast_results_to_pairwise_16s_iden_dict(blastn_output, align_len_cutoff, cov_cutoff):

    pairwise_iden_dict = {}
    for match in open(blastn_output):
        match_split = match.strip().split('\t')
        query = match_split[0]
        subject = match_split[1]
        iden = float(match_split[2])
        align_len = int(match_split[3])
        query_len = int(match_split[12])
        subject_len = int(match_split[13])
        coverage_q = float(align_len) * 100 / float(query_len)
        coverage_s = float(align_len) * 100 / float(subject_len)

        if (align_len >= align_len_cutoff) and (query != subject) and (coverage_q >= cov_cutoff) and (coverage_s >= cov_cutoff):
            query_to_subject_key = '__|__'.join(sorted([query, subject]))
            if query_to_subject_key not in pairwise_iden_dict:
                pairwise_iden_dict[query_to_subject_key] = iden
            else:
                if iden > pairwise_iden_dict[query_to_subject_key]:
                    pairwise_iden_dict[query_to_subject_key] = iden

    return pairwise_iden_dict


def get_GapFilling_stats_by_assembly(free_living_16s_ref_file,
                                     free_living_ctg_ref_file,
                                     mini_assembly_to_16s_reads,
                                     mini_assembly_to_ctg_reads,
                                     ctg_level_min_link,
                                     mini_assembly_to_16s_ctg_connector,
                                     gnm_to_ctg_connector,
                                     marker_to_ctg_gnm_Key_connector,
                                     max_within_cate_diff_pct,
                                     max_between_cate_diff_pct,
                                     stats_GapFilling_ctg,
                                     stats_GapFilling_gnm):

    round2_free_living_16s_ref_dict = {}
    for free_living_read_16s in open(free_living_16s_ref_file):
        free_living_read_16s_split = free_living_read_16s.strip().split('\t')
        if len(free_living_read_16s_split) > 1:
            read_16s_id = free_living_read_16s_split[0]
            read_16s_refs = free_living_read_16s_split[1].split(',')
            round2_free_living_16s_ref_dict[read_16s_id] = read_16s_refs

    round2_free_living_ctg_ref_dict = {}
    for free_living_read_ctg in open(free_living_ctg_ref_file):
        free_living_read_ctg_split = free_living_read_ctg.strip().split('\t')
        read_ctg_id = free_living_read_ctg_split[0]
        read_ctg_refs = free_living_read_ctg_split[1].split(',')
        read_ctg_refs_no_suffix = []
        for each_read_ctg_ref in read_ctg_refs:
            if each_read_ctg_ref[-2:] in ['_l', '_r']:
                each_read_ctg_ref_no_suffix = each_read_ctg_ref[:-2]
                read_ctg_refs_no_suffix.append(each_read_ctg_ref_no_suffix)
        round2_free_living_ctg_ref_dict[read_ctg_id] = read_ctg_refs_no_suffix

    mini_assembly_to_16s_dict = {}
    mini_assembly_to_16s_linking_reads_dict = {}
    for each_mini_assembly in open(mini_assembly_to_16s_reads):
        mini_assembly_split = each_mini_assembly.strip().split('\t')
        mini_assembly_id = mini_assembly_split[0]
        mini_assembly_mapped_reads = mini_assembly_split[1].split(',')
        for each_mapped_read in mini_assembly_mapped_reads:
            mapped_read_16s_refs = round2_free_living_16s_ref_dict.get(each_mapped_read, [])
            for each_mapped_read_16s_ref in mapped_read_16s_refs:
                mini_assembly_to_16s_key = '%s%s%s' % (mini_assembly_id, mini_assembly_to_16s_ctg_connector, each_mapped_read_16s_ref)
                if mini_assembly_to_16s_key not in mini_assembly_to_16s_dict:
                    mini_assembly_to_16s_dict[mini_assembly_to_16s_key] = 1
                    mini_assembly_to_16s_linking_reads_dict[mini_assembly_to_16s_key] = {each_mapped_read}
                else:
                    mini_assembly_to_16s_dict[mini_assembly_to_16s_key] += 1
                    mini_assembly_to_16s_linking_reads_dict[mini_assembly_to_16s_key].add(each_mapped_read)

    mini_assembly_to_ctg_dict = {}
    mini_assembly_to_ctg_linking_reads_dict = {}
    for each_mini_assembly in open(mini_assembly_to_ctg_reads):
        mini_assembly_split = each_mini_assembly.strip().split('\t')
        mini_assembly_id = mini_assembly_split[0]
        mini_assembly_mapped_reads = mini_assembly_split[1].split(',')
        for each_mapped_read in mini_assembly_mapped_reads:
            mapped_read_ctg_refs = round2_free_living_ctg_ref_dict.get(each_mapped_read, [])
            for each_mapped_read_ctg_ref in mapped_read_ctg_refs:
                mini_assembly_to_ctg_key = '%s%s%s' % (mini_assembly_id, mini_assembly_to_16s_ctg_connector, each_mapped_read_ctg_ref)
                if mini_assembly_to_ctg_key not in mini_assembly_to_ctg_dict:
                    mini_assembly_to_ctg_dict[mini_assembly_to_ctg_key] = 1
                    mini_assembly_to_ctg_linking_reads_dict[mini_assembly_to_ctg_key] = {each_mapped_read}
                else:
                    mini_assembly_to_ctg_dict[mini_assembly_to_ctg_key] += 1
                    mini_assembly_to_ctg_linking_reads_dict[mini_assembly_to_ctg_key].add(each_mapped_read)


    mini_assembly_to_16s_dict_reformatted = {}
    max_link_nun_dict_16s = {}
    for each in mini_assembly_to_16s_dict:
        mini_assembly_id = each.split(mini_assembly_to_16s_ctg_connector)[0]
        seq_16s_id = each.split(mini_assembly_to_16s_ctg_connector)[1]
        linkage_num = mini_assembly_to_16s_dict[each]
        seq_16s_with_num = '%s__num__%s' % (seq_16s_id, linkage_num)
        if linkage_num >= ctg_level_min_link:

            # add to mini_assembly_to_16s_dict_reformatted
            if mini_assembly_id not in mini_assembly_to_16s_dict_reformatted:
                mini_assembly_to_16s_dict_reformatted[mini_assembly_id] = {seq_16s_with_num}
            else:
                mini_assembly_to_16s_dict_reformatted[mini_assembly_id].add(seq_16s_with_num)

            # add to max_link_nun_dict_16s
            if seq_16s_id not in max_link_nun_dict_16s:
                max_link_nun_dict_16s[seq_16s_id] = linkage_num
            else:
                if linkage_num > max_link_nun_dict_16s[seq_16s_id]:
                    max_link_nun_dict_16s[seq_16s_id] = linkage_num

    print('mini_assembly_to_16s_dict_reformatted: %s' % mini_assembly_to_16s_dict_reformatted)

    mini_assembly_to_ctg_dict_reformatted = {}
    max_link_nun_dict_ctg = {}
    for each in mini_assembly_to_ctg_dict:
        mini_assembly_id = each.split(mini_assembly_to_16s_ctg_connector)[0]
        ctg_id = each.split(mini_assembly_to_16s_ctg_connector)[1]
        linkage_num = mini_assembly_to_ctg_dict[each]
        ctg_with_num = '%s__num__%s' % (ctg_id, linkage_num)
        if linkage_num >= ctg_level_min_link:

            # add to mini_assembly_to_ctg_dict_reformatted
            if mini_assembly_id not in mini_assembly_to_ctg_dict_reformatted:
                mini_assembly_to_ctg_dict_reformatted[mini_assembly_id] = {ctg_with_num}
            else:
                mini_assembly_to_ctg_dict_reformatted[mini_assembly_id].add(ctg_with_num)

            # add to max_link_nun_dict_ctg
            if ctg_id not in max_link_nun_dict_ctg:
                max_link_nun_dict_ctg[ctg_id] = linkage_num
            else:
                if linkage_num > max_link_nun_dict_ctg[ctg_id]:
                    max_link_nun_dict_ctg[ctg_id] = linkage_num

    #print(mini_assembly_to_ctg_dict)
    #print(mini_assembly_to_ctg_dict_reformatted)

    mini_assembly_linked_both = set(mini_assembly_to_16s_dict_reformatted).intersection(mini_assembly_to_ctg_dict_reformatted)
    print('mini_assembly_linked_both: %s' % mini_assembly_linked_both)

    stats_GapFilling_ctg_handle = open(stats_GapFilling_ctg, 'w')
    stats_GapFilling_gnm_dict = {}
    for each_mini_assembly in mini_assembly_linked_both:
        linked_16s = mini_assembly_to_16s_dict_reformatted[each_mini_assembly]
        linked_ctg = mini_assembly_to_ctg_dict_reformatted[each_mini_assembly]
        linked_16s_num_list = [int(i.split('__num__')[1]) for i in linked_16s]
        linked_ctg_num_list = [int(i.split('__num__')[1]) for i in linked_ctg]
        linked_16s_num_max = max(linked_16s_num_list)
        linked_ctg_num_max = max(linked_ctg_num_list)

        if (min(linked_16s_num_max, linked_ctg_num_max) * 100 / max(linked_16s_num_max, linked_ctg_num_max)) >= max_between_cate_diff_pct:

            linked_16s_filtered = [i for i in linked_16s if int(i.split('__num__')[1])*100/linked_16s_num_max >= max_within_cate_diff_pct]
            linked_ctg_filtered = [i for i in linked_ctg if int(i.split('__num__')[1])*100/linked_ctg_num_max >= max_within_cate_diff_pct]

            for each_linked_16s in linked_16s_filtered:
                linked_16s_id = each_linked_16s.split('__num__')[0]
                linked_16s_num = int(each_linked_16s.split('__num__')[1])
                linked_16s_num_pct_by_max = linked_16s_num * 100 / max_link_nun_dict_16s[linked_16s_id]

                for each_linked_ctg in linked_ctg_filtered:
                    linked_ctg_id = each_linked_ctg.split('__num__')[0]
                    linked_gnm_id = linked_ctg_id.split(gnm_to_ctg_connector)[0]
                    linked_ctg_num = int(each_linked_ctg.split('__num__')[1])
                    linked_ctg_num_pct_by_max = linked_ctg_num*100/max_link_nun_dict_ctg[linked_ctg_id]

                    if (linked_16s_num_pct_by_max >= 50) and (linked_ctg_num_pct_by_max >= 50):
                        stats_GapFilling_ctg_handle.write('%s\t%s\t%s\n' % (linked_16s_id, linked_ctg_id, (linked_16s_num + linked_ctg_num)))
                        marker_to_gnm_key = '%s%s%s' % (linked_16s_id, marker_to_ctg_gnm_Key_connector, linked_gnm_id)
                        if marker_to_gnm_key not in stats_GapFilling_gnm_dict:
                            stats_GapFilling_gnm_dict[marker_to_gnm_key] = (linked_16s_num + linked_ctg_num)
                        else:
                            stats_GapFilling_gnm_dict[marker_to_gnm_key] += (linked_16s_num + linked_ctg_num)
    stats_GapFilling_ctg_handle.close()

    stats_GapFilling_gnm_handle = open(stats_GapFilling_gnm, 'w')
    stats_GapFilling_gnm_handle.write('MarkerGene,GenomicSeq,Number\n')
    for each_16s_to_gnm in stats_GapFilling_gnm_dict:
        each_16s_to_gnm_split = each_16s_to_gnm.split(marker_to_ctg_gnm_Key_connector)
        id_16s = each_16s_to_gnm_split[0]
        id_gnm = each_16s_to_gnm_split[1]
        linkage_num = stats_GapFilling_gnm_dict[each_16s_to_gnm]
        stats_GapFilling_gnm_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (id_16s, id_gnm, linkage_num))
    stats_GapFilling_gnm_handle.close()


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def sort_csv_by_col(file_in, file_out, col_header):
    df_in        = pd.read_csv(file_in)
    df_in_sorted = df_in.sort_values(by=[col_header], ascending=False)
    df_in_sorted.to_csv(file_out, index=False)


def filter_linkages_iteratively(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict, marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff, min_linkages, min_linkages_for_uniq_linked_16s, within_gnm_linkage_num_diff, file_out):

    # get MarkerGene_to_GenomicSeq_dict
    MarkerGene_to_GenomicSeq_dict = {}
    for each_linkage in open(file_in):
        if not each_linkage.startswith('MarkerGene,GenomicSeq,Number'):
            each_linkage_split = each_linkage.strip().split(',')
            MarkerGene_id = each_linkage_split[0][12:]
            GenomicSeq_id = each_linkage_split[1][12:]
            linkage_num = int(each_linkage_split[2])
            if linkage_num > 1:
                if MarkerGene_id not in MarkerGene_to_GenomicSeq_dict:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id] = {GenomicSeq_id}
                else:
                    MarkerGene_to_GenomicSeq_dict[MarkerGene_id].add(GenomicSeq_id)

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    gnm_max_link_num_dict = {}
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    GenomicSeq_best_marker_dict = {}
    for each_match in open(file_in_sorted):
        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])

            current_min_linkage = min_linkages_for_uniq_linked_16s
            if MarkerGene in MarkerGene_to_GenomicSeq_dict:
                if len(MarkerGene_to_GenomicSeq_dict[MarkerGene]) > 1:
                    current_min_linkage = min_linkages

            if linkage_num >= current_min_linkage:
                if MarkerGene not in MarkerGene_with_assignment:

                    # consider depth
                    if min_16s_gnm_multiple > 0:

                        # get marker and genome depth
                        MarkerGene_depth = marker_gene_depth_dict.get(MarkerGene, 'na')
                        GenomicSeq_depth = genomic_seq_depth_dict.get(GenomicSeq, 'na')
                        marker_genome_depth_ratio = 'na'
                        if (MarkerGene_depth != 'na') and (GenomicSeq_depth != 'na'):
                            if GenomicSeq_depth > 0:
                                marker_genome_depth_ratio = MarkerGene_depth / GenomicSeq_depth

                        if marker_genome_depth_ratio >= min_16s_gnm_multiple:
                            if GenomicSeq not in GenomicSeq_best_marker_dict:
                                GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                                gnm_max_link_num_dict[GenomicSeq] = linkage_num
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
                            else:
                                # get identity with best marker
                                current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                                key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))
                                iden_with_best_marker = pairwise_16s_iden_dict.get(key_str, 0)
                                if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                    gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                                    if (linkage_num * 100 / gnm_max_link_num) >= within_gnm_linkage_num_diff:
                                        file_out_handle.write(each_match)
                                        MarkerGene_with_assignment.add(MarkerGene)
                                    else:
                                        MarkerGene_with_assignment.add(MarkerGene)
                    # ignore depth
                    else:
                        if GenomicSeq not in GenomicSeq_best_marker_dict:
                            GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                            gnm_max_link_num_dict[GenomicSeq] = linkage_num
                            file_out_handle.write(each_match)
                            MarkerGene_with_assignment.add(MarkerGene)
                        else:
                            # get identity with best marker
                            current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                            key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))
                            iden_with_best_marker = pairwise_16s_iden_dict.get(key_str, 0)
                            if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                gnm_max_link_num = gnm_max_link_num_dict[GenomicSeq]
                                if (linkage_num*100/gnm_max_link_num) >= within_gnm_linkage_num_diff:
                                    file_out_handle.write(each_match)
                                    MarkerGene_with_assignment.add(MarkerGene)
                                else:
                                    MarkerGene_with_assignment.add(MarkerGene)
    file_out_handle.close()


########################################################################################################################

wd                                              = '/Users/songweizhi/Desktop/stats_mini_link'
blast_results_all_vs_all_16s                    = '%s/MBARC26_0516_16S_all_vs_all_blastn.tab'   % wd
free_living_16s_ref_file                        = '%s/round2_free_living_16s_refs.txt'          % wd
free_living_ctg_ref_file                        = '%s/round2_free_living_ctg_refs.txt'          % wd
mini_assembly_to_16s_reads                      = '%s/mini_assembly_to_16s_reads.txt'           % wd
mini_assembly_to_ctg_reads                      = '%s/mini_assembly_to_ctg_reads.txt'           % wd
min_aln_16s                                     = 500
min_cov_16s                                     = 30
min_iden_16s                                    = 98
ctg_level_min_link                              = 3
mini_assembly_to_16s_ctg_connector              = '___Mini___'
gnm_to_ctg_connector                            = '___C___'
marker_to_ctg_gnm_Key_connector                 = '___M___'
within_gnm_linkage_num_diff                     = 80
max_mini_assembly_link_num_diff_between_ctg_16s = 10
mean_depth_dict_gnm                             = {}
mean_depth_dict_16s                             = {}
min_16s_gnm_multiple                            = 0
min_link_num                                    = 8

# file out
stats_GapFilling_ctg                            = '%s/stats_GapFilling_ctg.txt'                 % wd
stats_GapFilling_file                           = '%s/stats_GapFilling_gnm.txt'                 % wd
stats_GapFilling_file_filtered                  = '%s/stats_GapFilling_gnm_filtered.txt'        % wd


########################################################################################################################

get_GapFilling_stats_by_assembly(free_living_16s_ref_file,
                                 free_living_ctg_ref_file,
                                 mini_assembly_to_16s_reads,
                                 mini_assembly_to_ctg_reads,
                                 ctg_level_min_link,
                                 mini_assembly_to_16s_ctg_connector,
                                 gnm_to_ctg_connector,
                                 marker_to_ctg_gnm_Key_connector,
                                 within_gnm_linkage_num_diff,
                                 max_mini_assembly_link_num_diff_between_ctg_16s,
                                 stats_GapFilling_ctg,
                                 stats_GapFilling_file)

pairwise_16s_iden_dict = blast_results_to_pairwise_16s_iden_dict(blast_results_all_vs_all_16s, min_aln_16s, min_cov_16s)

filter_linkages_iteratively(stats_GapFilling_file, 'Number', pairwise_16s_iden_dict, mean_depth_dict_gnm,
                            mean_depth_dict_16s, min_16s_gnm_multiple, min_iden_16s, min_link_num, min_link_num,
                            within_gnm_linkage_num_diff, stats_GapFilling_file_filtered)




#!/usr/bin/env python3

import os
import argparse
from Bio import SeqIO

'''
Example command
python3 get_gene_location_from_gbk.py -h
python3 get_gene_location_from_gbk.py -id gene_id.txt -gbk Test_pcofg_prodigal_output -o gene_location.txt
'''

parser = argparse.ArgumentParser()
parser.add_argument('-id',  required=True, help='gene id file, one id per line')
parser.add_argument('-gbk', required=True, help='folder holds gbk files, e.g. prodigal_output folder from MetaCHIP')
parser.add_argument('-o',   required=True, help='output file')
args = vars(parser.parse_args())

gene_id_file        = args['id']
gbk_folder          = args['gbk']
gene_location_file  = args['o']


genome_to_gene_dict = dict()
for each_gene in open(gene_id_file):
    gene_id = each_gene.strip()
    genome_id = '_'.join(gene_id.split('_')[:-1])
    if genome_id not in genome_to_gene_dict:
        genome_to_gene_dict[genome_id] = {gene_id}
    else:
        genome_to_gene_dict[genome_id].add(gene_id)


gene_location_file_handle = open(gene_location_file, 'w')
gene_location_file_handle.write('Gene\tGenome\tContig\n' )
for each_gnm in genome_to_gene_dict:

    gnm_gbk = '%s/%s.gbk' % (gbk_folder, each_gnm)
    current_gnm_genes = genome_to_gene_dict[each_gnm]

    if os.path.isfile(gnm_gbk) is False:
        for each_current_gene in current_gnm_genes:
            gene_location_file_handle.write('%s\t%s\tgbk_not_found\n' % (each_current_gene, each_gnm))
    else:
        gene_to_ctg_dict = dict()
        for ctg_record in SeqIO.parse(gnm_gbk, "genbank"):
            ctg_record_id = ctg_record.id
            for gene_record in ctg_record.features:
                if 'locus_tag' in gene_record.qualifiers:
                    gene_record_id = gene_record.qualifiers['locus_tag'][0]
                    if gene_record_id in current_gnm_genes:
                        gene_to_ctg_dict[gene_record_id] = ctg_record_id
        # write out
        for each_gene_id in current_gnm_genes:
            gene_location_file_handle.write('%s\t%s\t%s\n' % (each_gene_id, each_gnm, gene_to_ctg_dict.get(each_gene_id, 'gene_not_found_in_gbk')))

gene_location_file_handle.close()



import os
import glob
import shutil
import argparse
import pandas as pd
from Bio import SeqIO
from time import sleep
import multiprocessing as mp
from datetime import datetime
from distutils.spawn import find_executable


def extract_reads_worker(argument_list):

    reads_file_in    = argument_list[0]
    reads_to_extract = argument_list[1]
    reads_file_out   = argument_list[2]

    reads_file_out_handle = open(reads_file_out, 'w')
    for read_record in SeqIO.parse(reads_file_in, 'fasta'):
        if read_record.id in reads_to_extract:
            reads_file_out_handle.write('>%s\n' % read_record.id)
            reads_file_out_handle.write('%s\n' % read_record.seq)
    reads_file_out_handle.close()


def split_list(list_in, subset_num):

    list_in_formatted = [i for i in list_in]

    # get the number of element per subset
    file_num_per_folder = round(len(list_in_formatted) / subset_num)

    n = 1
    lol_out = []
    while n <= subset_num:

        if n < subset_num:
            current_subset_elements = {i for i in list_in_formatted[(file_num_per_folder * (n - 1)):(file_num_per_folder * n)]}
            lol_out.append(current_subset_elements)
        else:
            current_subset_elements = {i for i in list_in_formatted[(file_num_per_folder * (n - 1)):]}
            lol_out.append(current_subset_elements)

        n += 1

    return lol_out


def extracted_reads_with_multiprocessing(reads_r1, reads_r2, r1_to_extract, r2_to_extract, output_folder, num_threads):
    solely_perfectly_mapped_reads_r1_splitted = split_list(r1_to_extract, num_threads // 2)
    solely_perfectly_mapped_reads_r2_splitted = split_list(r2_to_extract, num_threads // 2)

    argument_list_for_extract_reads_worker = []
    extract_reads_file_index_r1 = 1
    for reads_subset_r1 in solely_perfectly_mapped_reads_r1_splitted:
        current_output_file = '%s/extract_r1_subset_%s.fasta' % (output_folder, extract_reads_file_index_r1)
        argument_list_for_extract_reads_worker.append([reads_r1, reads_subset_r1, current_output_file])
        extract_reads_file_index_r1 += 1

    extract_reads_file_index_r2 = 1
    for reads_subset_r2 in solely_perfectly_mapped_reads_r2_splitted:
        current_output_file = '%s/extract_r2_subset_%s.fasta' % (output_folder, extract_reads_file_index_r2)
        argument_list_for_extract_reads_worker.append([reads_r2, reads_subset_r2, current_output_file])
        extract_reads_file_index_r2 += 1

    # extract reads with multiprocessing
    pool = mp.Pool(processes=num_threads)
    pool.map(extract_reads_worker, argument_list_for_extract_reads_worker)
    pool.close()
    pool.join()


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if each_element.isalpha() is True:
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def get_free_living_mate(ref_in, reads_r1, reads_r2, end_seq_len, minCigarM, max_gap_to_end, bowtie_build_exe, bowtie2_exe, num_threads):

    ref_in_path, ref_in_basename, ref_in_ext = sep_path_basename_ext(ref_in)

    ref_subset = '%s/%s_ends_%sbp%s'   % (ref_in_path, ref_in_basename, end_seq_len, ref_in_ext)
    sam_file   = '%s/%s_ends_%sbp.sam' % (ref_in_path, ref_in_basename, end_seq_len)

    # get ref seqs subset
    ref_subset_len_dict = {}
    ref_subset_handle = open(ref_subset, 'w')
    for ref_seq in SeqIO.parse(ref_in, 'fasta'):
        ref_seq_id = ref_seq.id
        ref_seq_len = len(ref_seq.seq)
        if ref_seq_len < end_seq_len*3:
            ref_subset_handle.write('>%s\n' % ref_seq_id)
            ref_subset_handle.write('%s\n' % ref_seq.seq)
            ref_subset_len_dict[ref_seq_id] = ref_seq_len
        else:
            ref_seq_left_end_id     = '%s_l' % ref_seq_id
            ref_seq_right_end_id    = '%s_r' % ref_seq_id
            ref_seq_left_end        = ref_seq.seq[:end_seq_len]
            ref_seq_right_end       = ref_seq.seq[-end_seq_len:]

            # write out left end
            ref_subset_handle.write('>%s\n' % ref_seq_left_end_id)
            ref_subset_handle.write('%s\n' % ref_seq_left_end)

            # write out right end
            ref_subset_handle.write('>%s\n' % ref_seq_right_end_id)
            ref_subset_handle.write('%s\n' % ref_seq_right_end)
    ref_subset_handle.close()

    # index ref seq subset
    ref_subset_no_ext = '.'.join(ref_subset.split('.')[:-1])
    bowtie2_index_ref_cmd = '%s -f %s %s --quiet --threads %s' % (bowtie_build_exe, ref_subset, ref_subset_no_ext, num_threads)
    os.system(bowtie2_index_ref_cmd)

    # mapping
    bowtie2_mapping_cmd = '%s -x %s -1 %s -2 %s -S %s -f --local --no-unal --quiet --threads %s' % (bowtie2_exe, ref_subset_no_ext, reads_r1, reads_r2, sam_file, num_threads)
    os.system(bowtie2_mapping_cmd)

    # parse sam file
    all_mapped_reads = set()
    qualified_reads_dict = {}
    qualified_reads_to_ref_dict = {}
    for each_line in open(sam_file):
        if not each_line.startswith('@'):
            each_line_split = each_line.strip().split('\t')
            read_id         = each_line_split[0]
            read_id_base    = '.'.join(read_id.split('.')[:-1])
            read_strand     = read_id.split('.')[-1]
            ref_id          = each_line_split[2]
            ref_pos         = int(each_line_split[3])
            cigar           = each_line_split[5]
            cigar_splitted  = cigar_splitter(cigar)

            all_mapped_reads.add(read_id)

            qualified_mapping = False
            if ref_id[-2:] == '_l':
                if ref_pos <= max_gap_to_end:
                    if (len(cigar_splitted) == 1) and (cigar[-1] == 'M') and(int(cigar[:-1]) >= minCigarM):
                        qualified_mapping = True

                    elif (len(cigar_splitted) == 2) and (cigar_splitted[-1][-1] == 'M') and (cigar_splitted[0][-1] == 'S') and (int(cigar_splitted[-1][:-1]) >= minCigarM) and (ref_pos == 1):
                        qualified_mapping = True

            elif ref_id[-2:] == '_r':

                if (len(cigar_splitted) == 1) and (cigar[-1] == 'M') and (int(cigar[:-1]) >= minCigarM):
                    ref_pos_end = ref_pos + int(cigar[:-1])

                    if (end_seq_len - ref_pos_end) <= max_gap_to_end:
                        qualified_mapping = True

                elif (len(cigar_splitted) == 2) and (cigar_splitted[0][-1] == 'M') and (cigar_splitted[1][-1] == 'S') and (int(cigar_splitted[0][:-1]) >= minCigarM):
                    if (ref_pos + int(cigar_splitted[0][:-1]) - 1) == end_seq_len:
                        qualified_mapping = True

            else:
                ref_len = ref_subset_len_dict[ref_id]
                if (len(cigar_splitted) == 1) and (cigar[-1] == 'M') and (int(cigar[:-1]) >= minCigarM):
                    ref_pos_end = ref_pos + int(cigar[:-1])

                    # left side
                    if ref_pos <= max_gap_to_end:
                        qualified_mapping = True

                    # right side
                    elif (ref_len - ref_pos_end) <= max_gap_to_end:
                        qualified_mapping = True

                if len(cigar_splitted) == 2:

                    # left side
                    if (cigar_splitted[-1][-1] == 'M') and (cigar_splitted[0][-1] == 'S') and (int(cigar_splitted[-1][:-1]) >= minCigarM) and (ref_pos == 1):
                        qualified_mapping = True

                    # right side
                    elif (cigar_splitted[-0][-1] == 'M') and (cigar_splitted[1][-1] == 'S') and (int(cigar_splitted[0][:-1]) >= minCigarM):
                        if (ref_pos + int(cigar_splitted[0][:-1]) - 1) == ref_len:
                            qualified_mapping = True

            if qualified_mapping is True:
                qualified_reads_to_ref_dict[read_id] = ref_id

                if read_id_base not in qualified_reads_dict:
                    qualified_reads_dict[read_id_base] = [read_strand]
                else:
                    qualified_reads_dict[read_id_base].append(read_strand)


    reads_to_extract_to_ref_dict = {}
    for qualified_read in qualified_reads_dict:
        read_strand = qualified_reads_dict[qualified_read]
        if len(read_strand) == 1:

            mapped_mate = ''
            mate_to_extract = ''
            if read_strand == ['1']:
                mapped_mate     = '%s.1' % (qualified_read)
                mate_to_extract = '%s.2' % (qualified_read)
            if read_strand == ['2']:
                mapped_mate     = '%s.2' % (qualified_read)
                mate_to_extract = '%s.1' % (qualified_read)

            if mate_to_extract not in all_mapped_reads:
                reads_to_extract_to_ref_dict[mate_to_extract] = qualified_reads_to_ref_dict[mapped_mate]

    return reads_to_extract_to_ref_dict


########################################################################################################################

wd = '/Users/songweizhi/Desktop/777'

marker_gene_seqs_1st_round_unlinked    = '%s/rest_16S.fasta' % wd
combined_1st_round_unlinked_mags              = '%s/rest_mags.fna' % wd
reads_file_r1       = '%s/MBARC26_R1_0.05.fasta'    % wd
reads_file_r2       = '%s/MBARC26_R2_0.05.fasta'    % wd
end_seq_len         = 3000
minCigarM           = 50
max_gap_to_end      = 100
num_threads         = 4

pwd_bowtie2_build_exe    = '/Users/songweizhi/Softwares/bowtie2/bowtie2-build'
pwd_bowtie2_exe         = '/Users/songweizhi/Softwares/bowtie2/bowtie2'
pwd_spades_exe          = '/Users/songweizhi/Softwares/SPAdes-3.14.1/bin/spades.py'

pwd_bowtie2_build_exe    = 'bowtie2-build'
pwd_bowtie2_exe         = 'bowtie2'
pwd_spades_exe          = 'spades.py'


# define output
free_living_mate_gnm        = '%s/free_living_mate_ctg.txt'             % wd
free_living_mate_16s        = '%s/free_living_mate_16s.txt'             % wd
extracted_reads_folder      = '%s/free_living_reads'                    % wd
extracted_reads_cbd         = '%s/free_living_read_combined.fasta'      % wd
spades_wd                   = '%s/combined_free_living_reads_SPAdes_wd' % wd
spades_assemblies           = '%s/scaffolds.fasta'                      % spades_wd
sam_file                    = '%s/scaffolds.sam'                        % wd


########################################################################################################################

reads_to_extract_to_ref_dict_gnm = get_free_living_mate(combined_1st_round_unlinked_mags, reads_file_r1, reads_file_r2, end_seq_len, minCigarM, max_gap_to_end, pwd_bowtie2_build_exe, pwd_bowtie2_exe, num_threads)
reads_to_extract_to_ref_dict_16s = get_free_living_mate(marker_gene_seqs_1st_round_unlinked, reads_file_r1, reads_file_r2, end_seq_len, minCigarM, max_gap_to_end, pwd_bowtie2_build_exe, pwd_bowtie2_exe, num_threads)

# write out free_living_mate_gnm
free_living_mate_gnm_handle = open(free_living_mate_gnm, 'w')
for fl_read_gnm in reads_to_extract_to_ref_dict_gnm:
    free_living_mate_gnm_handle.write('%s\t%s\n' % (fl_read_gnm, reads_to_extract_to_ref_dict_gnm[fl_read_gnm]))
free_living_mate_gnm_handle.close()

# write out free_living_mate_16s
free_living_mate_16s_handle = open(free_living_mate_16s, 'w')
for fl_read_16s in reads_to_extract_to_ref_dict_16s:
    free_living_mate_16s_handle.write('%s\t%s\n' % (fl_read_16s, reads_to_extract_to_ref_dict_16s[fl_read_16s]))
free_living_mate_16s_handle.close()

# reads_to_extract_to_ref_dict_gnm = {}
# for i in open(free_living_mate_gnm):
#     i_split = i.strip().split('\t')
#     reads_to_extract_to_ref_dict_gnm[i_split[0]] = i_split[1]
#
# reads_to_extract_to_ref_dict_16s = {}
# for j in open(free_living_mate_16s):
#     j_split = j.strip().split('\t')
#     reads_to_extract_to_ref_dict_16s[j_split[0]] = j_split[1]


####################################################  extract reads ####################################################

os.mkdir(extracted_reads_folder)

extract_list_gnm = set()
extract_list_16s = set()
extract_list_combined_r1 = set()
extract_list_combined_r2 = set()
for fl_mate_gnm in reads_to_extract_to_ref_dict_gnm:
    extract_list_gnm.add(fl_mate_gnm)
    if fl_mate_gnm[-2:] == '.1':
        extract_list_combined_r1.add(fl_mate_gnm)
    if fl_mate_gnm[-2:] == '.2':
        extract_list_combined_r2.add(fl_mate_gnm)
for fl_mate_16s in reads_to_extract_to_ref_dict_16s:
    extract_list_16s.add(fl_mate_16s)
    if fl_mate_16s[-2:] == '.1':
        extract_list_combined_r1.add(fl_mate_16s)
    if fl_mate_16s[-2:] == '.2':
        extract_list_combined_r2.add(fl_mate_16s)


# remove overlap
extract_list_combined_r1_no_overlap = set()
for r1 in extract_list_combined_r1:
    if (r1 in extract_list_gnm) and (r1 in extract_list_16s):
        pass
    else:
        extract_list_combined_r1_no_overlap.add(r1)

extract_list_combined_r2_no_overlap = set()
for r2 in extract_list_combined_r2:
    if (r2 in extract_list_gnm) and (r2 in extract_list_16s):
        pass
    else:
        extract_list_combined_r2_no_overlap.add(r2)

# extract reads with multiprocessing
extracted_reads_with_multiprocessing(reads_file_r1, reads_file_r2, extract_list_combined_r1_no_overlap, extract_list_combined_r2_no_overlap, extracted_reads_folder, num_threads)

# combine extracted reads
os.system('cat %s/*.fasta > %s' % (extracted_reads_folder, extracted_reads_cbd))
os.system('rm -r %s' % extracted_reads_folder)


####################################################### assemble #######################################################

spades_cmd = '%s -s %s -o %s -t %s -k 21,33,55,75,99,127 --only-assembler' % (pwd_spades_exe, extracted_reads_cbd, spades_wd, num_threads)
os.system(spades_cmd)


######################################################## mapping #######################################################

spades_assemblies_no_ext = '.'.join(spades_assemblies.split('.')[:-1])
index_ref_cmd = '%s -f %s %s' % (pwd_bowtie2_build_exe, spades_assemblies, spades_assemblies_no_ext)
mapping_cmd = '%s -x %s -U %s -S %s -p %s -f' % (pwd_bowtie2_exe, spades_assemblies_no_ext, extracted_reads_cbd, sam_file, num_threads)
os.system(index_ref_cmd)
os.system(mapping_cmd)


#################################################### parse sam file ####################################################

short_ctg_to_reads_dict = {}
for each_line in open(sam_file):
    if not each_line.startswith('@'):
        each_line_split = each_line.strip().split('\t')
        read_id         = each_line_split[0]
        read_id_base    = '.'.join(read_id.split('.')[:-1])
        read_strand     = read_id.split('.')[-1]
        ref_id          = each_line_split[2]
        ref_pos         = int(each_line_split[3])
        cigar           = each_line_split[5]
        cigar_splitted  = cigar_splitter(cigar)
        if (len(cigar_splitted) == 1) and (cigar[-1] == 'M'):
            if ref_id not in short_ctg_to_reads_dict:
                short_ctg_to_reads_dict[ref_id] = [read_id]
            else:
                short_ctg_to_reads_dict[ref_id].append(read_id)


for short_ctg in short_ctg_to_reads_dict:
    short_ctg_mapped_reads = short_ctg_to_reads_dict[short_ctg]
    for mapped_read in short_ctg_mapped_reads:

        mapped_read_mate_ref = ''
        if mapped_read in reads_to_extract_to_ref_dict_gnm:
            mapped_read_mate_ref = reads_to_extract_to_ref_dict_gnm[mapped_read]
        if mapped_read in reads_to_extract_to_ref_dict_16s:
            mapped_read_mate_ref = reads_to_extract_to_ref_dict_16s[mapped_read]

        print('%s\t%s' % (short_ctg, mapped_read_mate_ref))

    print()

'''
export PATH=/Users/songweizhi/Softwares/bowtie2:$PATH
export PATH=/Users/songweizhi/Softwares/samtools-1.11/bin:$PATH
cd /Users/songweizhi/Desktop/777
cat free_living_read_16s.fasta free_living_read_gnm.fasta > free_living_read_cbd.fasta
bowtie2-build -f combined_refs.fa combined_refs
bowtie2 -x combined_refs -U free_living_read_16s.fasta -S free_living_read_16s.sam -p 4 -f
bowtie2 -x combined_refs -U free_living_read_gnm.fasta -S free_living_read_gnm.sam -p 4 -f
bowtie2 -x combined_refs -U free_living_read_cbd.fasta -S free_living_read_cbd.sam -p 4 -f
samtools view -bS free_living_read_16s.sam -o free_living_read_16s.bam
samtools view -bS free_living_read_gnm.sam -o free_living_read_gnm.bam
samtools view -bS free_living_read_cbd.sam -o free_living_read_cbd.bam
samtools sort free_living_read_16s.bam -o free_living_read_16s_sorted.bam
samtools sort free_living_read_gnm.bam -o free_living_read_gnm_sorted.bam
samtools sort free_living_read_cbd.bam -o free_living_read_cbd_sorted.bam
samtools index free_living_read_16s_sorted.bam
samtools index free_living_read_gnm_sorted.bam
samtools index free_living_read_cbd_sorted.bam

module load spades/3.14.0
cd /srv/scratch/z5039045/MarkerMAG_wd/new_algorithm
spades.py -s free_living_read_cbd.fasta -o free_living_read_cbd_spades -t 4 --only-assembler
cd /srv/scratch/z5039045/MarkerMAG_wd/new_algorithm/free_living_read_cbd_spades
cd /Users/songweizhi/Desktop/777
bowtie2-build -f scaffolds.fasta scaffolds
bowtie2 -x scaffolds -U free_living_read_cbd.fasta -S scaffolds.sam -p 4 -f
samtools view -bS scaffolds.sam -o scaffolds.bam
samtools sort scaffolds.bam -o scaffolds_sorted.bam
samtools index scaffolds_sorted.bam

cd /Users/songweizhi/Desktop/777
/Users/songweizhi/Softwares/SPAdes-3.14.1/bin/spades.py -s combined_free_living_reads.fasta -o combined_free_living_reads_spades_wd -t 4 --only-assembler

# on Katana
module load python/3.7.3
source ~/mypython3env/bin/activate
module load spades/3.14.0
module load bowtie/2.3.5.1
cd /srv/scratch/z5039045/MarkerMAG_wd/new_algorithm
python3 tmp_6_2nd_round.py

'''



import os
import glob
from Bio import SeqIO

'''


MBARC26_mis_1_10_5_combined_linkages.txt	Genome	17/23(73.91)	17/17(100.0)	Wrong:	    Unlinked:AS,CG,CP,HR,PS,TR
MBARC26_mis_2_10_5_combined_linkages.txt	Genome	17/23(73.91)	17/17(100.0)	Wrong:	    Unlinked:AS,CG,CP,HR,PS,TR
MBARC26_mis_3_10_5_combined_linkages.txt	Genome	16/23(69.57)	16/16(100.0)	Wrong:	    Unlinked:AS,CG,CP,EC,HR,PS,TR
MBARC26_mis_4_10_5_combined_linkages.txt	Genome	18/23(78.26)	18/18(100.0)	Wrong:	    Unlinked:CG,CP,HR,PS,TR
MBARC26_mis_5_10_5_combined_linkages.txt	Genome	17/23(73.91)	17/17(100.0)	Wrong:	    Unlinked:AS,CG,CP,HR,PS,TR
MBARC26_mis_6_10_5_combined_linkages.txt	Genome	16/23(69.57)	16/17(94.12)	Wrong:EC	Unlinked:AS,CG,CP,HR,PS,TR
MBARC26_mis_7_10_5_combined_linkages.txt	Genome	17/23(73.91)	17/17(100.0)	Wrong:	    Unlinked:CG,CP,EC,HR,PS,TR
MBARC26_mis_8_10_5_combined_linkages.txt	Genome	17/23(73.91)	17/17(100.0)	Wrong:	    Unlinked:CG,CP,EC,HR,PS,TR
MBARC26_mis_9_10_5_combined_linkages.txt	Genome	17/23(73.91)	17/17(100.0)	Wrong:	    Unlinked:CG,CP,EC,HR,PS,TR

MBARC26_mis_3_10_5_spades_combined_linkages.txt	Genome	17/23(73.91)	17/17(100.0)	Wrong:	Unlinked:AS,CG,CP,EC,PS,TR
MBARC26_mis_4_10_5_spades_combined_linkages.txt	Genome	18/23(78.26)	18/18(100.0)	Wrong:	Unlinked:CG,CP,EC,PS,TR
MBARC26_mis_5_10_5_spades_combined_linkages.txt	Genome	18/23(78.26)	18/18(100.0)	Wrong:	Unlinked:CG,CP,EC,PS,TR

'''

########################################################################################################################

wd                          = '/Users/songweizhi/Desktop/MarkerMAG_wd/2_MBARC26'
marker_gene_seqs            = '%s/MBARC26_all_depth_assemblies_uclust_0.999_99.9_500bp.fasta'   % wd
mag_folder                  = '%s/Refined_refined_bins_renamed'                                 % wd
mag_file_extension          = 'fna'
combined_linkage_file       = '%s/MBARC26_mis_6_10_5_combined_linkages.txt'                          % wd

'''

# SB and AS formed one bin,

[2021-04-07 22:01:23] MBARC26_0407_very_sensitive	Genome	19/23(82.61)	90.48	Unrecovered(4):AS,CP,CA,PS
[2021-04-07 21:49:24] MBARC26_0407_sensitive	    Genome	19/23(82.61)	90.48	Unrecovered(4):AS,CP,CA,PS
[2021-04-07 21:58:59] MBARC26_0407_default	        Genome	21/23(91.3)	    95.45	Unrecovered(2):AS,CP
[2021-04-07 21:53:56] MBARC26_0407_specific	        Genome	21/23(91.3)	    95.45	Unrecovered(2):AS,CP
[2021-04-07 21:50:10] MBARC26_0407_very_specific	Genome	20/23(86.96)	95.24	Unrecovered(3):AS,CP,TR
[2021-04-07 21:54:39] MBARC26_0407_super_specific	Genome	19/23(82.61)	95.0	Unrecovered(4):AS,CP,HR,TR

'''

########################################################################################################################

marker_id_set = set()
for marker_seq_record in SeqIO.parse(marker_gene_seqs, 'fasta'):
    marker_id_set.add(marker_seq_record.id)

linkage_file_basename = os.path.basename(combined_linkage_file)

mag_file_re             = '%s/*%s' % (mag_folder, mag_file_extension)
mag_file_list           = [os.path.basename(file_name) for file_name in glob.glob(mag_file_re)]
mag_file_list_no_ext    = {'.'.join(i.split('.')[:-1]) for i in mag_file_list}

markers_with_wrong_linkage = set()
markers_with_right_linkage = set()
genomes_with_right_linkage = set()
genomes_with_wrong_linkage = set()
all_linked_genomes = set()
total_linkages = 0
right_linkages = 0
for each_match in open(combined_linkage_file):
    if not each_match.startswith('MarkerGene\tGenomicSeq\tLinkage\tStep'):
        total_linkages += 1
        match_split = each_match.strip().split('\t')
        linkage_num = int(match_split[2])
        MarkerGene_genome = match_split[0].split('_')[0]
        GenomicSeq_genome = match_split[1]
        all_linked_genomes.add(GenomicSeq_genome)
        if MarkerGene_genome == GenomicSeq_genome:
            right_linkages += 1
            genomes_with_right_linkage.add(GenomicSeq_genome)
            markers_with_right_linkage.add(match_split[0])
        else:
            genomes_with_wrong_linkage.add(GenomicSeq_genome)
            markers_with_wrong_linkage.add(match_split[0])

recovery_by_marker_str      = '%s/%s(%s)'               % (len(markers_with_right_linkage), len(marker_id_set), float("{0:.2f}".format(len(markers_with_right_linkage)*100/len(marker_id_set))))
accuracy_by_marker_str      = '%s/%s(%s)'               % (right_linkages, total_linkages, float("{0:.2f}".format(right_linkages*100/total_linkages)))
unrecovered_by_marker_str   = 'Wrong:%s\tUnlinked:'     % (','.join(markers_with_wrong_linkage))
recovery_by_genome_str      = '%s/%s(%s)'               % (len(genomes_with_right_linkage), len(mag_file_list_no_ext), float("{0:.2f}".format(len(genomes_with_right_linkage)*100/len(mag_file_list_no_ext))))
accuracy_by_genome_str      = '%s/%s(%s)'               % (len(genomes_with_right_linkage), len(all_linked_genomes), float("{0:.2f}".format(len(genomes_with_right_linkage)*100/len(all_linked_genomes))))
unrecovered_by_genome_str   = 'Wrong:%s\tUnlinked:%s'   % (','.join(genomes_with_wrong_linkage), ','.join(sorted([i for i in (mag_file_list_no_ext - genomes_with_right_linkage - genomes_with_wrong_linkage)])))

print('Linkages\tBy\tRecovery\tAccuracy\tUnrecovered')
print('%s\tMarker\t%s\t%s\t%s' % (linkage_file_basename, recovery_by_marker_str, accuracy_by_marker_str, unrecovered_by_marker_str))
print('%s\tGenome\t%s\t%s\t%s' % (linkage_file_basename, recovery_by_genome_str, accuracy_by_genome_str, unrecovered_by_genome_str))




def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if each_element.isalpha() is True:
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


clipping_reads_not_matched_part_seq = '/Users/songweizhi/Desktop/clipping_not_matched_part_seq.fasta'
pwd_samfile                         = '/Users/songweizhi/Desktop/BH_ER_050417_assembled_16S_uclust_0.995.sam'
#pwd_samfile                         = '/Users/songweizhi/Desktop/3_GI_assembled_16S_uclust_0.995.sam'

perfect_match_min_cigar_M_pct       = 90
perfect_match_max_cigar_S_pct       = 10


# export clipping mapped reads and perfectly mapped reads
all_mapped_reads_set = set()
clipping_mapped_reads_list = set()
clipping_reads_mapped_part_dict = {}
perfectly_mapped_reads_dict = {}
clipping_reads_not_matched_part_seq_handle = open(clipping_reads_not_matched_part_seq, 'w')
for each_read in open(pwd_samfile):

    if not each_read.startswith('@'):
        each_read_split = each_read.strip().split('\t')
        read_id = each_read_split[0]
        read_id_base = '.'.join(read_id.split('.')[:-1])
        read_strand = read_id.split('.')[-1]
        ref_id = each_read_split[2]
        ref_id_with_prefix = 'MarkerGene__%s' % each_read_split[2]
        ref_pos = int(each_read_split[3])
        cigar = each_read_split[5]
        read_seq = each_read_split[9]
        cigar_splitted = cigar_splitter(cigar)
        read_id_with_ref_pos = '%s__x__%s__x__%s' % (read_id, ref_id, ref_pos)
        all_mapped_reads_set.add(read_id)

        # for perfectly mapped reads
        if ('M' in cigar) and (len(cigar_splitted) == 1):
            if read_id_base not in perfectly_mapped_reads_dict:
                perfectly_mapped_reads_dict[read_id_base] = {read_strand: [ref_id_with_prefix]}
            else:
                if read_strand not in perfectly_mapped_reads_dict[read_id_base]:
                    perfectly_mapped_reads_dict[read_id_base][read_strand] = [ref_id_with_prefix]
                else:
                    perfectly_mapped_reads_dict[read_id_base][read_strand].append(ref_id_with_prefix)

        # for clipping mapped reads
        if ('S' in cigar) and (len(cigar_splitted) == 2):
            cigar_M_len = 0
            cigar_S_len = 0
            if cigar_splitted[0][-1] == 'M':
                cigar_M_len = int(cigar_splitted[0][:-1])
                cigar_S_len = int(cigar_splitted[1][:-1])
            if cigar_splitted[1][-1] == 'M':
                cigar_M_len = int(cigar_splitted[1][:-1])
                cigar_S_len = int(cigar_splitted[0][:-1])

            cigar_M_pct = cigar_M_len * 100 / (cigar_M_len + cigar_S_len)
            cigar_S_pct = cigar_S_len * 100 / (cigar_M_len + cigar_S_len)

            # for clipping reads with unmapped part >= min_cigar_S
            if (cigar_M_pct >= 0.3) and (cigar_S_pct >= 0.3):
                read_seq_left = read_seq[: int(cigar_splitted[0][:-1])]
                read_seq_right = read_seq[-int(cigar_splitted[1][:-1]):]

                if cigar_splitted[0][-1] == 'M':

                    # write out the sequence of unmapped part
                    clipping_reads_not_matched_part_seq_handle.write('>%s_r\n' % read_id_with_ref_pos)
                    clipping_reads_not_matched_part_seq_handle.write(read_seq_right + '\n')

                    # store the match info of mapped part
                    if ('%s_l' % read_id_with_ref_pos) not in clipping_reads_mapped_part_dict:
                        clipping_reads_mapped_part_dict[('%s_l' % read_id_with_ref_pos)] = [ref_id_with_prefix]
                    else:
                        clipping_reads_mapped_part_dict[('%s_l' % read_id_with_ref_pos)].append(ref_id_with_prefix)

                if cigar_splitted[1][-1] == 'M':

                    # write out the sequence of unmapped part
                    clipping_reads_not_matched_part_seq_handle.write('>%s_l\n' % read_id_with_ref_pos)
                    clipping_reads_not_matched_part_seq_handle.write(read_seq_left + '\n')

                    # store the match info of mapped part
                    if ('%s_r' % read_id_with_ref_pos) not in clipping_reads_mapped_part_dict:
                        clipping_reads_mapped_part_dict[('%s_r' % read_id_with_ref_pos)] = [ref_id_with_prefix]
                    else:
                        clipping_reads_mapped_part_dict[('%s_r' % read_id_with_ref_pos)].append(ref_id_with_prefix)

                clipping_mapped_reads_list.add(read_id)

            # for clipping reads with unmapped part < min_cigar_S
            # treat these reads as perfectly mapped reads
            if (cigar_M_pct >= perfect_match_min_cigar_M_pct) and (cigar_S_pct < perfect_match_max_cigar_S_pct):
                if read_id_base not in perfectly_mapped_reads_dict:
                    perfectly_mapped_reads_dict[read_id_base] = {read_strand: [ref_id_with_prefix]}
                else:
                    if read_strand not in perfectly_mapped_reads_dict[read_id_base]:
                        perfectly_mapped_reads_dict[read_id_base][read_strand] = [ref_id_with_prefix]
                    else:
                        perfectly_mapped_reads_dict[read_id_base][read_strand].append(ref_id_with_prefix)

clipping_reads_not_matched_part_seq_handle.close()


########################################## extract reads with multiprocessing ##########################################

perfectly_mapped_read_singleton_dict = {}
for perfectly_mapped_read in perfectly_mapped_reads_dict:
    current_value = perfectly_mapped_reads_dict[perfectly_mapped_read]
    if len(current_value) == 1:
        perfectly_mapped_read_singleton_dict[perfectly_mapped_read] = current_value

# get the id of paired reads to extract
solely_perfectly_mapped_reads_r1 = set()
solely_perfectly_mapped_reads_r2 = set()
m = 0
n = 0
for perfectly_mapped_read in perfectly_mapped_read_singleton_dict:
    current_value = perfectly_mapped_read_singleton_dict[perfectly_mapped_read]
    strand = list(current_value.keys())[0]
    if strand == '1':
        r2_to_extract = '%s.2' % perfectly_mapped_read
        if r2_to_extract not in all_mapped_reads_set:
            solely_perfectly_mapped_reads_r2.add(r2_to_extract)
    if strand == '2':
        r1_to_extract = '%s.1' % perfectly_mapped_read
        if r1_to_extract not in all_mapped_reads_set:
            solely_perfectly_mapped_reads_r1.add(r1_to_extract)


#print('all_mapped_reads:\t\t%s'                 % len(all_mapped_reads_set))
print('clipping_mapped_reads:\t%s'              % len(clipping_mapped_reads_list))
print('fully_mapped_reads:\t%s'             % len(perfectly_mapped_reads_dict))
print('fully_mapped_read_singleton:\t%s'    % len(perfectly_mapped_read_singleton_dict))
print('solely_fully_mapped_reads_r1:\t%s'   % len(solely_perfectly_mapped_reads_r1))
print('solely_fullymapped_reads_r2:\t%s'   % len(solely_perfectly_mapped_reads_r2))

# print('m:\t%s' % m)
# print('n:\t%s' % n)

#print(perfectly_mapped_reads_dict)
#print(perfectly_mapped_read_singleton_dict)

'''
                                Kelp    GI
clipping_mapped_reads:	        25062   153285
fully_mapped_reads:	            17001   264100
fully_mapped_read_singleton:	4173    70535
solely_fully_mapped_reads_r1:	251     11223
solely_fullymapped_reads_r2:	383     26215

GI_R1.fastq         27.98 Gbp
GI_R2.fastq         21.84 Gbp
Spades assemblies:  412.6 Mbp

BH_ER_050417_R1.fastq   24.51 Gbp
BH_ER_050417_R2.fastq   23.19 Gbp
Spades assemblies:      1.12 Gbp

'''


#!/usr/bin/env python3

# Copyright (C) 2020, Weizhi Song, Torsten Thomas.
# songwz03@gmail.com or t.thomas@unsw.edu.au

# MarkerMAG is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# MarkerMAG is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import argparse
from Bio import SeqIO


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


parser = argparse.ArgumentParser()

parser.add_argument('-m',       required=True,                              help='Matam assemblies')
parser.add_argument('-c',       required=True,                              help='Control, e.g. Barrnap predicted 16S rRNA gene sequences')
parser.add_argument('-i',       required=False, type=float, default=99.5,   help='Identity cutoff, default: 99.5')
parser.add_argument('-l',       required=False, type=int,   default=1500,   help='Alignment length cutoff, default: 1500')
parser.add_argument('-noblast', required=False, action="store_true",        help='Skip blastn')

args = vars(parser.parse_args())
matam_16s_seqs   = args['m']
barrnap_16s_seqs = args['c']
iden_cutoff      = args['i']
aln_len_cutoff   = args['l']
skip_blastn      = args['noblast']


matam_16s_seqs_file_path, matam_16s_seqs_file_basename, matam_16s_seqs_file_extension = sep_path_basename_ext(matam_16s_seqs)
matam_16s_blastn = '%s/%s_%s_%sbp_blastn.tab' % (matam_16s_seqs_file_path, matam_16s_seqs_file_basename, iden_cutoff, aln_len_cutoff)
output_file      = '%s/%s_%s_%sbp.fasta' % (matam_16s_seqs_file_path, matam_16s_seqs_file_basename, iden_cutoff, aln_len_cutoff)
output_file_txt  = '%s/%s_%s_%sbp.txt'   % (matam_16s_seqs_file_path, matam_16s_seqs_file_basename, iden_cutoff, aln_len_cutoff)

blast_parameters = '-evalue 1e-5 -outfmt "6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore qlen slen" -task blastn -num_threads 1'
blast_cmd = 'blastn -query %s -subject %s -out %s %s' % (matam_16s_seqs, barrnap_16s_seqs, matam_16s_blastn, blast_parameters)
if skip_blastn is False:
    os.system(blast_cmd)


# get the number of control sequences
genome_id_set = set()
control_seq_id_list = set()
for seq_record in SeqIO.parse(barrnap_16s_seqs, 'fasta'):
    control_seq_id_list.add(seq_record.id)
    genome_id_set.add(seq_record.id.split('_16S_')[0])


query_to_subject_dict = {}
query_to_subject_genome_dict = {}
subject_genome_to_query_dict = {}
matched_matam_set = set()
matched_control_set = set()
matched_control_set_genome_level = set()
for match in open(matam_16s_blastn):
    match_split = match.strip().split('\t')
    query = match_split[0]
    subject = match_split[1]
    subject_genome = subject.split('_16S_')[0]
    iden = float(match_split[2])
    aln_len = int(match_split[3])
    query_len = int(match_split[12])
    subject_len = int(match_split[13])
    coverage_q = float(aln_len) * 100 / float(query_len)
    coverage_s = float(aln_len) * 100 / float(subject_len)

    if (iden >= iden_cutoff) and (aln_len > aln_len_cutoff):

        # get query_to_subject_dict
        if query not in query_to_subject_dict:
            query_to_subject_dict[query] = {subject}
        else:
            query_to_subject_dict[query].add(subject)

        # get query_to_subject_genome_dict
        if query not in query_to_subject_genome_dict:
            query_to_subject_genome_dict[query] = {subject_genome}
        else:
            query_to_subject_genome_dict[query].add(subject_genome)

        # get subject_genome_to_query_dict
        if subject_genome not in subject_genome_to_query_dict:
            subject_genome_to_query_dict[subject_genome] = {query}
        else:
            subject_genome_to_query_dict[subject_genome].add(query)

        matched_matam_set.add(query)
        matched_control_set.add(subject)
        matched_control_set_genome_level.add(subject_genome)


multiple_assign = False
for i in query_to_subject_genome_dict:
    if len(query_to_subject_genome_dict[i]) > 1:
        multiple_assign = True


# get rename dict
rename_dict = {}
for genome in subject_genome_to_query_dict:
    n = 1
    for m16s in subject_genome_to_query_dict[genome]:
        if m16s not in rename_dict:
            rename_dict[m16s] = {'%s_m%s' % (genome, n)}
        else:
            rename_dict[m16s].add('%s_m%s' % (genome, n))
        n += 1


# rename and write out
output_file_handle = open(output_file, 'w')
for m16s_record in SeqIO.parse(matam_16s_seqs, 'fasta'):
    if m16s_record.id in rename_dict:
        new_name_list = rename_dict[m16s_record.id]
        for new_name in new_name_list:
            output_file_handle.write('>%s matam_%s\n' % (new_name, m16s_record.id) )
            output_file_handle.write('%s\n' % m16s_record.seq)
output_file_handle.close()


os.remove(matam_16s_blastn)


output_file_txt_handle = open(output_file_txt, 'w')
for matched_assembly in rename_dict:
    new_names = rename_dict[matched_assembly]
    matched_to_list = query_to_subject_dict[matched_assembly]
    for new_name in new_names:
        output_file_txt_handle.write('%s\t%s\n' % (new_name, ','.join(matched_to_list)))
output_file_txt_handle.close()


######################################################## report ########################################################

# get unrecovered genome
unrecovered_genome_list = []
for genome_id in genome_id_set:
    if genome_id not in matched_control_set_genome_level:
        unrecovered_genome_list.append(genome_id)

# get unrecovered control set
unrecovered_control_set = set()
for i in control_seq_id_list:
    if i not in matched_control_set:
        unrecovered_control_set.add(i)

unrecovered_str = 'Unrecovered(%s):%s' % (len(unrecovered_control_set), ','.join(sorted([i for i in unrecovered_control_set])))
unrecovered_str_genome_level = 'Unrecovered(%s):%s' % (len(unrecovered_genome_list), ','.join(sorted(unrecovered_genome_list)))

print('Iden\tLen\tControl\tMatam\tUnrecovered')
print('Marker\t%s%s\t%sbp\t%s/%s\t%s\t%s' % (iden_cutoff, '%', aln_len_cutoff, len(matched_control_set), len(control_seq_id_list), len(matched_matam_set), unrecovered_str))
print('Genome\t%s%s\t%sbp\t%s/%s\t%s\t%s' % (iden_cutoff, '%', aln_len_cutoff, len(matched_control_set_genome_level), len(genome_id_set), len(matched_matam_set), unrecovered_str_genome_level))

if multiple_assign is True:
    print('Some/one Matam assembly matched to multiple genomes!')
else:
    print('Good, no Matam assembly matched to multiple genomes!')


########################################################################################################################


'''
cd /Users/songweizhi/Desktop/111
python3 /Users/songweizhi/PycharmProjects/MarkerMAG/script_backup/rename_matam_assemblies.py -c /srv/scratch/z5039045/MarkerMAG_wd/genome_selection_3/combined_16S.ffn -m scaffolds.NR.min_500bp.abd.fa -i 99.9 -a 1300

python3 /srv/scratch/z5039045/MarkerMAG_wd/rename_matam_assemblies.py -c /srv/scratch/z5039045/MarkerMAG_wd/genome_selection_3/combined_16S.ffn -m scaffolds.NR.min_500bp.abd.fa -i 99.9 -a 1300

cd /Users/songweizhi/Desktop/999
python3 /Users/songweizhi/PycharmProjects/MarkerMAG/script_backup/rename_matam_assemblies.py -c combined_16S.ffn -m combined_all_depth_assemblies_iden99.5_uniq.fasta -i 99.5 -l 1300

'''



bin_to_ref_txt   = '/Users/songweizhi/Desktop/assess_linkages_hc/stats_bin_to_ref_imag99.5.txt'
ref_with_16s_txt = '/Users/songweizhi/Desktop/assess_linkages_hc/file_in/ref_genomes_hc_with_16S.txt'


refs_with_16s_set = set()
for each_ref in open(ref_with_16s_txt):
    refs_with_16s_set.add(each_ref.strip()[4:])


n = 0
for eac_line in open(bin_to_ref_txt):
    eac_line_split = eac_line.strip().split('\t')
    bin_id = eac_line_split[0]
    ref_list = eac_line_split[1].split(',')
    ref_list_onyl_id = {i[4:] for i in ref_list}
    # print(eac_line_split)
    # print(bin_id)
    # print(ref_list_onyl_id)
    # print()

    ref_with_16s = False
    for each_ref in ref_list_onyl_id:
        if each_ref in refs_with_16s_set:
            ref_with_16s = True

    if ref_with_16s is False:
        print(bin_id)
        n += 1

print(n)


'''
cami_hc_110
cami_hc_115
cami_hc_116
cami_hc_119
cami_hc_122
cami_hc_133
cami_hc_150
cami_hc_154
cami_hc_53
cami_hc_64
cami_hc_80
cami_hc_87
cami_hc_99
'''




gtdb_metadata      = '/Users/songweizhi/Desktop/MarkerMAG_wd/select_strain/bac120_metadata_r95.tsv'
get_genome_from    = ['p__Bacteroidota']
taxon_to_exculde   = []
complete_genome    = True
ncbi_type_material = True
gtdb_type_strain   = True


x = 0
qualified_genome_num = 0
for genome_record in open(gtdb_metadata):
    if not genome_record.startswith('accession	'):
        genome_record_split             = genome_record.strip().split('\t')
        accession                       = genome_record_split[0]
        gtdb_taxonomy                   = genome_record_split[16]
        ncbi_assembly_level             = genome_record_split[45]
        ncbi_type_material_designation  = genome_record_split[84]
        ssu_count                       = int(genome_record_split[91])

        if ('o__Enterobacterales' in gtdb_taxonomy) and (ssu_count == 6) and (ncbi_assembly_level not in ['Contig', 'Scaffold']) and (ncbi_type_material_designation == 'assembly from type material') and (genome_record_split[17] == 'type strain of species'):
            print('%s\t%s' % (accession, gtdb_taxonomy))
            qualified_genome_num += 1

print('Number of qualified genome is %s' % qualified_genome_num)










def paired_blast_results_to_dict(blastn_results, iden_cutoff, query_cov_cutoff):

    query_to_subject_list_dict = {}
    for blast_hit in open(blastn_results):
        blast_hit_split = blast_hit.strip().split('\t')
        query = blast_hit_split[0]
        subject = blast_hit_split[1]
        subject_with_prefix = 'GenomicSeq__%s' % subject
        iden = float(blast_hit_split[2])
        align_len = int(blast_hit_split[3])
        query_len = int(blast_hit_split[12])
        subject_len = int(blast_hit_split[13])
        coverage_q = float(align_len) * 100 / float(query_len)

        # for perfect hits
        if (iden >= iden_cutoff) and (coverage_q == 100):
            if query not in query_to_subject_list_dict:
                query_to_subject_list_dict[query] = [subject_with_prefix]
            else:
                query_to_subject_list_dict[query].append(subject_with_prefix)

        # for nearly perfect hits
        elif (iden >= iden_cutoff) and (query_cov_cutoff <= coverage_q < 100):
            s_l = sorted([int(blast_hit_split[8]), int(blast_hit_split[9])])[0]
            s_r = sorted([int(blast_hit_split[8]), int(blast_hit_split[9])])[1]
            subject_min_gap = min([s_l, (subject_len - s_r)])
            if subject_min_gap <= 5000:
                if query not in query_to_subject_list_dict:
                    query_to_subject_list_dict[query] = [subject_with_prefix]
                else:
                    query_to_subject_list_dict[query].append(subject_with_prefix)

    return query_to_subject_list_dict


blastn_results = '/Users/songweizhi/Desktop/unmapped_paired_reads_blast.txt'
reads_iden_cutoff = 100
reads_cov_cutoff = 90
query_to_sub_dict = paired_blast_results_to_dict(blastn_results, reads_iden_cutoff, reads_cov_cutoff)
print(query_to_sub_dict)
print(len(query_to_sub_dict))




from Bio import SeqIO


reads_0              = '/Users/songweizhi/Desktop/test_spades/rd2_read_to_extract_flanking_both_R12_up_failed.fa'
reads_1              = '/Users/songweizhi/Desktop/test_spades/rd2_read_to_extract_flanking_both_R12_up_succeed.fa'
shared_reads_from_r0 = '/Users/songweizhi/Desktop/test_spades/shared_reads_from_r0.fa'
shared_reads_from_r1 = '/Users/songweizhi/Desktop/test_spades/shared_reads_from_r1.fa'
uniq_reads_from_r0   = '/Users/songweizhi/Desktop/test_spades/uniq_reads_from_r0.fa'
uniq_reads_from_r1   = '/Users/songweizhi/Desktop/test_spades/uniq_reads_from_r1.fa'


reads_0_id_set = set()
for each in SeqIO.parse(reads_0, 'fasta'):
    reads_0_id_set.add(each.id)
reads_1_id_set = set()
for each in SeqIO.parse(reads_1, 'fasta'):
    reads_1_id_set.add(each.id)


shared_reads = set(reads_0_id_set).intersection(reads_1_id_set)


shared_reads_from_r0_handle = open(shared_reads_from_r0, 'w')
uniq_reads_from_r0_handle = open(uniq_reads_from_r0, 'w')
for each in SeqIO.parse(reads_0, 'fasta'):
    if each.id in shared_reads:
        shared_reads_from_r0_handle.write('>%s\n' % each.id)
        shared_reads_from_r0_handle.write('%s\n' % str(each.seq))
    else:
        uniq_reads_from_r0_handle.write('>%s\n' % each.id)
        uniq_reads_from_r0_handle.write('%s\n' % str(each.seq))
shared_reads_from_r0_handle.close()
uniq_reads_from_r0_handle.close()


shared_reads_from_r1_handle = open(shared_reads_from_r1, 'w')
uniq_reads_from_r1_handle = open(uniq_reads_from_r1, 'w')
for each in SeqIO.parse(reads_1, 'fasta'):
    if each.id in shared_reads:
        shared_reads_from_r1_handle.write('>%s\n' % each.id)
        shared_reads_from_r1_handle.write('%s\n' % str(each.seq))
    else:
        uniq_reads_from_r1_handle.write('>%s\n' % each.id)
        uniq_reads_from_r1_handle.write('%s\n' % str(each.seq))
shared_reads_from_r1_handle.close()
uniq_reads_from_r1_handle.close()


print(len(reads_0_id_set))
print(len(reads_1_id_set))
print(len(shared_reads))
print(len(reads_0_id_set) - len(shared_reads))
print(len(reads_1_id_set) - len(shared_reads))




import os
import glob
from Bio import SeqIO


genomic_assemblies          = ''
renamed_mag_folder          = '/Users/songweizhi/Desktop/777/Refined_refined_bins_renamed_renamed'
mag_file_extension          = 'fna'
marker_gene_seqs            = '/Users/songweizhi/Desktop/777/MBARC26_all_depth_assemblies.dereplicated_99.5_500bp.fasta'
link_stats_combined_table   = '/Users/songweizhi/Desktop/777/MBARC26_link_Matam_RealBins_500bp_depth_d0_mpl10_stats_combined_table.txt'
genomic_seq_type            = 'mag'

marker_gene_seqs_1st_round_unlinked = '/Users/songweizhi/Desktop/777/MBARC26_all_depth_assemblies.dereplicated_99.5_500bp_1st_round_unlinked.fasta'
combined_1st_round_unlinked_mags    = '/Users/songweizhi/Desktop/777/combined_1st_round_unlinked_mags.fasta'
combined_1st_round_unlinked_ctgs    = '/Users/songweizhi/Desktop/777/combined_1st_round_unlinked_ctgs.fasta'


# get linked marker genes and genomic sequences in 1st round
linked_marker_gene_set = set()
linked_genomic_seq_set = set()
for each_link in open(link_stats_combined_table):
    if not each_link.startswith('MarkerGene	GenomicSeq	Paired	Clipping'):
        each_link_split = each_link.strip().split('\t')
        linked_marker_gene_set.add(each_link_split[0])
        linked_genomic_seq_set.add(each_link_split[1])


# get the sequence of unlinked marker genes
marker_gene_seqs_1st_round_unlinked_handle = open(marker_gene_seqs_1st_round_unlinked, 'w')
for marker_gene_record in SeqIO.parse(marker_gene_seqs, 'fasta'):
    if marker_gene_record.id not in linked_marker_gene_set:
        marker_gene_seqs_1st_round_unlinked_handle.write('>%s\n' % marker_gene_record.id)
        marker_gene_seqs_1st_round_unlinked_handle.write('%s\n' % marker_gene_record.seq)
marker_gene_seqs_1st_round_unlinked_handle.close()


# get the sequence of unlinked genomic seqs
if genomic_seq_type == 'mag':

    # put all renamed mag into list
    renamed_gnm_re   = '%s/*.%s' % (renamed_mag_folder, mag_file_extension)
    renamed_gnm_list = [os.path.basename(file_name) for file_name in glob.glob(renamed_gnm_re)]
    renamed_gnm_list_no_ext = ['.'.join(i.split('.')[:-1]) for i in renamed_gnm_list]

    # keep only unlinked mags
    unlinked_mag_list_with_pwd = []
    for renamed_mag in renamed_gnm_list_no_ext:
        if renamed_mag not in linked_genomic_seq_set:
            pwd_renamed_mag = '%s/%s.%s' % (renamed_mag_folder, renamed_mag, mag_file_extension)
            unlinked_mag_list_with_pwd.append(pwd_renamed_mag)

    # combine unlinked mags
    cat_cmd = 'cat %s > %s' % (' '.join(unlinked_mag_list_with_pwd), combined_1st_round_unlinked_mags)
    os.system(cat_cmd)


# get the sequence of unlinked metagenomic assemblies
if genomic_seq_type == 'ctg':
    combined_1st_round_unlinked_ctgs_handle = open(combined_1st_round_unlinked_ctgs, 'w')
    for ctg_record in SeqIO.parse(genomic_assemblies, 'fasta'):
        if ctg_record.id not in linked_genomic_seq_set:
            SeqIO.write(ctg_record, combined_1st_round_unlinked_ctgs_handle, 'fasta')
    combined_1st_round_unlinked_ctgs_handle.close()





import os
import glob
import shutil
import argparse
import pandas as pd
from Bio import SeqIO
from time import sleep
import multiprocessing as mp
from datetime import datetime
from distutils.spawn import find_executable


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def get_max_value_in_dict(dict_in):
    max_value = 0
    max_value_key = ''
    for each_key in dict_in:
        if dict_in[each_key] > max_value:
            max_value = dict_in[each_key]
            max_value_key = each_key
    return max_value, max_value_key


def get_total_value_in_dict(dict_in):
    total_value = 0
    for each_key in dict_in:
        total_value += dict_in[each_key]
    return total_value


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if each_element.isalpha() is True:
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def sort_csv_by_col(file_in, file_out, col_header):
    df_in        = pd.read_csv(file_in)
    df_in_sorted = df_in.sort_values(by=[col_header], ascending=False)
    df_in_sorted.to_csv(file_out, index=False)


def blast_results_to_pairwise_16s_iden_dict(blastn_output, align_len_cutoff, cov_cutoff):

    pairwise_iden_dict = {}
    for match in open(blastn_output):
        match_split = match.strip().split('\t')
        query = match_split[0]
        subject = match_split[1]
        iden = float(match_split[2])
        align_len = int(match_split[3])
        query_len = int(match_split[12])
        subject_len = int(match_split[13])
        coverage_q = float(align_len) * 100 / float(query_len)
        coverage_s = float(align_len) * 100 / float(subject_len)

        if (align_len >= align_len_cutoff) and (query != subject) and (coverage_q >= cov_cutoff) and (coverage_s >= cov_cutoff):
            query_to_subject_key = '__|__'.join(sorted([query, subject]))
            if query_to_subject_key not in pairwise_iden_dict:
                pairwise_iden_dict[query_to_subject_key] = iden
            else:
                if iden > pairwise_iden_dict[query_to_subject_key]:
                    pairwise_iden_dict[query_to_subject_key] = iden

    return pairwise_iden_dict


def filter_linkages_iteratively(file_in, sort_by_col_header, pairwise_16s_iden_dict, genomic_seq_depth_dict, marker_gene_depth_dict, min_16s_gnm_multiple, within_genome_16s_divergence_cutoff, min_linkages, file_out):

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    file_out_handle = open(file_out, 'w')
    MarkerGene_with_assignment = set()
    GenomicSeq_best_marker_dict = {}
    for each_match in open(file_in_sorted):
        if each_match.startswith('MarkerGene,GenomicSeq,Number'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            MarkerGene = match_split[0][12:]
            GenomicSeq = match_split[1][12:]
            linkage_num = int(match_split[2])

            if linkage_num >= min_linkages:

                # consider depth
                if min_16s_gnm_multiple > 0:
                    MarkerGene_depth = marker_gene_depth_dict[MarkerGene]
                    GenomicSeq_depth = genomic_seq_depth_dict[GenomicSeq]
                    if (MarkerGene_depth/GenomicSeq_depth) >= min_16s_gnm_multiple:
                        if MarkerGene not in MarkerGene_with_assignment:

                            if GenomicSeq not in GenomicSeq_best_marker_dict:
                                GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)
                            else:
                                current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                                key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))

                                iden_with_best_marker = 0
                                if key_str in pairwise_16s_iden_dict:
                                    iden_with_best_marker = pairwise_16s_iden_dict[key_str]

                                if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                    file_out_handle.write(each_match)
                                    MarkerGene_with_assignment.add(MarkerGene)
                # ignore depth
                else:
                    if MarkerGene not in MarkerGene_with_assignment:
                        if GenomicSeq not in GenomicSeq_best_marker_dict:
                            GenomicSeq_best_marker_dict[GenomicSeq] = MarkerGene
                            file_out_handle.write(each_match)
                            MarkerGene_with_assignment.add(MarkerGene)
                        else:
                            current_GenomicSeq_best_marker = GenomicSeq_best_marker_dict[GenomicSeq]
                            key_str = '__|__'.join(sorted([MarkerGene, current_GenomicSeq_best_marker]))

                            iden_with_best_marker = 0
                            if key_str in pairwise_16s_iden_dict:
                                iden_with_best_marker = pairwise_16s_iden_dict[key_str]

                            if iden_with_best_marker >= within_genome_16s_divergence_cutoff:
                                file_out_handle.write(each_match)
                                MarkerGene_with_assignment.add(MarkerGene)

    file_out_handle.close()

    # remove tmp file
    # os.remove(file_in_sorted)


def filter_linkages_iteratively_simple(file_in, sort_by_col_header, min_linkages, file_out):

    file_in_path, file_in_basename, file_in_extension = sep_path_basename_ext(file_in)
    file_in_sorted = '%s/%s_sorted%s' % (file_in_path, file_in_basename, file_in_extension)

    # sort file in
    sort_csv_by_col(file_in, file_in_sorted, sort_by_col_header)

    # fileter linkage
    file_out_handle = open(file_out, 'w')
    gnm_ctg_16s_with_assignment = set()
    for each_match in open(file_in_sorted):
        if each_match.startswith('Gap_seq,'):
            file_out_handle.write(each_match)
        else:
            match_split = each_match.strip().split(',')
            gnm_ctg_16s_id = match_split[1]
            linkage_num = int(match_split[2])
            if (linkage_num >= min_linkages) and (gnm_ctg_16s_id not in gnm_ctg_16s_with_assignment):
                file_out_handle.write(each_match)
                gnm_ctg_16s_with_assignment.add(gnm_ctg_16s_id)
    file_out_handle.close()

    # remove tmp file
    # os.remove(file_in_sorted)


################################################### files/parameters ###################################################

wd = '/Users/songweizhi/Desktop/GapFilling_all'
wd = '/Users/songweizhi/Desktop/GapFilling_sep'

free_living_mate_gnm = '%s/free_living_mate_ctg.txt'                                        % wd
free_living_mate_16s = '%s/free_living_mate_16s.txt'                                        % wd
sam_file             = '%s/scaffolds.sam'                                                   % wd
gnm_ctg_connector    = '___'
min_read_num         = 10

if wd.endswith('_all'):
    blastdb_16s = '%s/MBARC26_all_depth_assemblies.dereplicated_99.5_500bp.fasta'           % wd
else:
    blastdb_16s = '%s/rest_16S.fasta'                                                       % wd


########################################################################################################################

# define output file name
blast_results_all_vs_all_16s   = '%s/16s_all_vs_all.txt'                                    % wd
stats_GapFilling_file          = '%s/stats_GapFilling.txt'                                  % wd
stats_GapFilling_file_filtered = '%s/stats_GapFilling_filtered.txt'                         % wd

stats_GapFilling_file_16s          = '%s/stats_GapFilling_16s.txt'                              % wd
stats_GapFilling_file_ctg          = '%s/stats_GapFilling_ctg.txt'                              % wd
stats_GapFilling_file_gnm          = '%s/stats_GapFilling_gnm.txt'                              % wd

stats_GapFilling_file_filtered_16s = '%s/stats_GapFilling_16s_filtered.txt'                         % wd
stats_GapFilling_file_filtered_ctg = '%s/stats_GapFilling_ctg_filtered.txt'                         % wd
stats_GapFilling_file_filtered_gnm = '%s/stats_GapFilling_gnm_filtered.txt'                         % wd


########################################################################################################################

reads_to_extract_to_ref_dict_gnm = {}
for i in open(free_living_mate_gnm):
    i_split = i.strip().split('\t')
    reads_to_extract_to_ref_dict_gnm[i_split[0]] = i_split[1]

reads_to_extract_to_ref_dict_16s = {}
for j in open(free_living_mate_16s):
    j_split = j.strip().split('\t')
    reads_to_extract_to_ref_dict_16s[j_split[0]] = j_split[1]

gap_seq_to_reads_dict = {}
for each_line in open(sam_file):
    if not each_line.startswith('@'):
        each_line_split = each_line.strip().split('\t')
        read_id         = each_line_split[0]
        read_id_base    = '.'.join(read_id.split('.')[:-1])
        read_strand     = read_id.split('.')[-1]
        ref_id          = each_line_split[2]
        ref_pos         = int(each_line_split[3])
        cigar           = each_line_split[5]
        cigar_splitted  = cigar_splitter(cigar)
        if (len(cigar_splitted) == 1) and (cigar[-1] == 'M'):
            if ref_id not in gap_seq_to_reads_dict:
                gap_seq_to_reads_dict[ref_id] = [read_id]
            else:
                gap_seq_to_reads_dict[ref_id].append(read_id)


stats_GapFilling_file_handle = open(stats_GapFilling_file, 'w')
stats_GapFilling_file_handle.write('MarkerGene,GenomicSeq,Number\n')
for short_ctg in gap_seq_to_reads_dict:
    short_ctg_mapped_reads = gap_seq_to_reads_dict[short_ctg]
    short_ctg_mapped_reads_mate_linkage_dict_16s = {}
    short_ctg_mapped_reads_mate_linkage_dict_ctg = {}
    marker_mate = False
    contig_mate = False
    for mapped_read in short_ctg_mapped_reads:

        mapped_read_mate_ref = ''
        if mapped_read in reads_to_extract_to_ref_dict_gnm:
            mapped_read_mate_ref = reads_to_extract_to_ref_dict_gnm[mapped_read]
        if mapped_read in reads_to_extract_to_ref_dict_16s:
            mapped_read_mate_ref = reads_to_extract_to_ref_dict_16s[mapped_read]

        # check mate mapped to marker gene or genomic sequence
        if '_m' in mapped_read_mate_ref:
            marker_mate = True

            if mapped_read_mate_ref not in short_ctg_mapped_reads_mate_linkage_dict_16s:
                short_ctg_mapped_reads_mate_linkage_dict_16s[mapped_read_mate_ref] = 1
            else:
                short_ctg_mapped_reads_mate_linkage_dict_16s[mapped_read_mate_ref] += 1

        mapped_read_mate_ref_genome = ''
        if 'NODE' in mapped_read_mate_ref:
            contig_mate = True
            mapped_read_mate_ref_genome = mapped_read_mate_ref.split(gnm_ctg_connector)[0]
            mapped_read_mate_ref = mapped_read_mate_ref_genome

            if mapped_read_mate_ref not in short_ctg_mapped_reads_mate_linkage_dict_ctg:
                short_ctg_mapped_reads_mate_linkage_dict_ctg[mapped_read_mate_ref] = 1
            else:
                short_ctg_mapped_reads_mate_linkage_dict_ctg[mapped_read_mate_ref] += 1

    if (len(short_ctg_mapped_reads_mate_linkage_dict_ctg) > 0) and (len(short_ctg_mapped_reads_mate_linkage_dict_16s) > 0):
        print(short_ctg)
        print(short_ctg_mapped_reads_mate_linkage_dict_ctg)
        print(short_ctg_mapped_reads_mate_linkage_dict_16s)
        print()
        for each_16s in short_ctg_mapped_reads_mate_linkage_dict_16s:
            num_16s = short_ctg_mapped_reads_mate_linkage_dict_16s[each_16s]
            for each_ctg in short_ctg_mapped_reads_mate_linkage_dict_ctg:
                num_ctg = short_ctg_mapped_reads_mate_linkage_dict_ctg[each_ctg]

                if (num_16s >= min_read_num) and (num_ctg >= min_read_num):
                    stats_GapFilling_file_handle.write('MarkerGene__%s,GenomicSeq__%s,%s\n' % (each_16s, each_ctg, (num_16s + num_ctg)))

stats_GapFilling_file_handle.close()


############################################# get pairwise_16s_iden_dict ##############################################

pwd_makeblastdb_exe = 'makeblastdb'
pwd_blastn_exe = 'blastn'
num_threads = 4
blast_parameters = '-evalue 1e-5 -outfmt "6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore qlen slen" -task blastn -num_threads %s' % num_threads

# makeblastdn with marker gene sequences
makeblastdb_16s_cmd = '%s -in %s -dbtype nucl -parse_seqids -logfile /dev/null' % (pwd_makeblastdb_exe, blastdb_16s)
os.system(makeblastdb_16s_cmd)

all_vs_all_16s_blastn_cmd = '%s -query %s -db %s -out %s %s' % (pwd_blastn_exe, blastdb_16s, blastdb_16s, blast_results_all_vs_all_16s, blast_parameters)
os.system(all_vs_all_16s_blastn_cmd)

pairwise_16s_iden_dict = blast_results_to_pairwise_16s_iden_dict(blast_results_all_vs_all_16s, 500, 90)


#######################################################################################################################

filter_linkages_iteratively(stats_GapFilling_file, 'Number', pairwise_16s_iden_dict, {}, {}, 0, 98, 0, stats_GapFilling_file_filtered)



import os
from Bio import AlignIO
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Align import MultipleSeqAlignment


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '.'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def mview_linkage(seq_file, plot_title, mafft_exe, mview_exe):

    # define file name
    seq_path, seq_basename, seq_ext = sep_path_basename_ext(seq_file)
    msa_file                    = '%s/%s.aln'               % (seq_path, seq_basename)
    msa_file_mview              = '%s/%s_MView.aln'         % (seq_path, seq_basename)
    msa_file_mviewd_html_tmp    = '%s/%s_MView_tmp.html'    % (seq_path, seq_basename)
    msa_file_mviewd_html        = '%s/%s_MView.html'        % (seq_path, seq_basename)

    gap_char                = ' '
    break_line_char         = '='

    # align sequences
    mafft_cmd = '%s --quiet --retree 1 %s > %s' % (mafft_exe, seq_file, msa_file)
    os.system(mafft_cmd)

    mapped_reads_dict = {}
    ref_id = ''
    ref_seq = ''
    current_line = 0
    for each_seq in AlignIO.read(msa_file, "fasta"):
        seq_id = each_seq.id
        seq_seq = str(each_seq.seq).upper()
        if current_line == 0:
            ref_id = seq_id
            ref_seq = seq_seq
        else:
            seq_id_base = '.'.join(seq_id.split('.')[:-1])
            seq_id_strand = seq_id.split('.')[-1]
            if seq_id_base not in mapped_reads_dict:
                if seq_id_strand == '1':
                    mapped_reads_dict[seq_id_base] = [seq_seq, '']
                if seq_id_strand == '2':
                    mapped_reads_dict[seq_id_base] = ['', seq_seq]
            else:
                if seq_id_strand == '1':
                    mapped_reads_dict[seq_id_base][0] = seq_seq
                if seq_id_strand == '2':
                    mapped_reads_dict[seq_id_base][1] = seq_seq

        current_line += 1

    # create an empty list to hold all sequences in a msa
    seq_record_list = []

    # add ref_seq to seq_record_list
    align_record_mview = MultipleSeqAlignment([])
    ref_seq_split_by_n = ref_seq.split('N')
    ref_seq_split_by_n_updated = []
    for segment in ref_seq_split_by_n:
        if ('-' in segment) and (segment == '-' * len(segment)):
            segment = 'N' * len(segment)
        ref_seq_split_by_n_updated.append(segment)
    ref_seq_updated = 'N'.join(ref_seq_split_by_n_updated)
    seq_record_list.append(SeqRecord(Seq(ref_seq_updated), id=ref_id, description='Reference'))

    # add break line
    seq_record_list.append(SeqRecord(Seq(break_line_char*len(ref_seq_updated)), id='#', description=''))

    # add paired reads to seq_record_list
    singleton_dict = {}
    overlapping_reads_dict = {}
    for each_read_base in mapped_reads_dict:

        r1_seq = mapped_reads_dict[each_read_base][0]
        r2_seq = mapped_reads_dict[each_read_base][1]

        if (r1_seq == '') or (r2_seq == ''):
            if not ((r1_seq == '') and (r2_seq == '')):
                singleton_dict[each_read_base] = mapped_reads_dict[each_read_base]
        else:
            overlapping_bps = 0
            for bp1, bp2 in zip(r1_seq, r2_seq):
                if (bp1 != '-') and (bp2 != '-'):
                    overlapping_bps += 1

            if overlapping_bps > 0:
                overlapping_reads_dict[each_read_base] = mapped_reads_dict[each_read_base]
            else:
                merge_r1_r2 = ''
                for bp1, bp2 in zip(r1_seq, r2_seq):
                    if (bp1 == '-') and (bp2 == '-'):
                        merge_r1_r2 += '-'
                    if (bp1 != '-') and (bp2 == '-'):
                        merge_r1_r2 += bp1
                    if (bp1 == '-') and (bp2 != '-'):
                        merge_r1_r2 += bp2
                seq_record_list.append(SeqRecord(Seq(merge_r1_r2), id=each_read_base, description='In_Pair'))

    # add break line
    seq_record_list.append(SeqRecord(Seq(break_line_char*len(ref_seq_updated)), id='#', description=''))

    # add overlapped reads to seq_record_list
    for each_overlapping_reads in overlapping_reads_dict:
        r1_id  = '%s.1' % each_overlapping_reads
        r2_id  = '%s.2' % each_overlapping_reads
        r1_seq = overlapping_reads_dict[each_overlapping_reads][0]
        r2_seq = overlapping_reads_dict[each_overlapping_reads][1]
        seq_record_list.append(SeqRecord(Seq(r1_seq), id=r1_id, description='Overlapped'))
        seq_record_list.append(SeqRecord(Seq(r2_seq), id=r2_id, description='Overlapped'))

    # add break line
    seq_record_list.append(SeqRecord(Seq(break_line_char*len(ref_seq_updated)), id='#', description=''))

    # add singleton to seq_record_list
    for each_singleton in singleton_dict:
        r1_id  = '%s.1' % each_singleton
        r2_id  = '%s.2' % each_singleton
        r1_seq = singleton_dict[each_singleton][0]
        r2_seq = singleton_dict[each_singleton][1]
        if (r1_seq != '') and (r2_seq == ''):
            seq_record_list.append(SeqRecord(Seq(r1_seq), id=r1_id, description='Unpaired'))
        if (r1_seq == '') and (r2_seq != ''):
            seq_record_list.append(SeqRecord(Seq(r2_seq), id=r2_id, description='Unpaired'))

    # add break line
    seq_record_list.append(SeqRecord(Seq(break_line_char*len(ref_seq_updated)), id='#', description=''))

    # get updated msa record
    align_record_mview = MultipleSeqAlignment(seq_record_list)

    # write out updated msa
    msa_file_mview_handle = open(msa_file_mview, 'w')
    AlignIO.write(align_record_mview, msa_file_mview_handle, 'fasta')
    msa_file_mview_handle.close()

    # visualize update msa
    # coloring: any,identity,mismatch,consensus,group
    mview_parameter_str = '-in fasta -moltype dna -colormap CLUSTAL_NUC -coloring any -css on -html head -ruler off -label0 -label4 -label5 -gap "%s"' % gap_char
    mview_cmd = '%s %s -title %s %s > %s' % (mview_exe, mview_parameter_str, plot_title, msa_file_mview, msa_file_mviewd_html_tmp)
    os.system(mview_cmd)

    msa_file_mviewd_html_handle = open(msa_file_mviewd_html, 'w')
    for each_line in open(msa_file_mviewd_html_tmp):
        if not (('Reference sequence' in each_line) or ('Colored by' in each_line)):
            msa_file_mviewd_html_handle.write(each_line)

    msa_file_mviewd_html_handle.close()


combined_ref_reads  = '/Users/songweizhi/Desktop/vis_msa/combined.fa'
msa_file            = '/Users/songweizhi/Desktop/vis_msa/combined.aln'
mview_exe           = '/Users/songweizhi/bin/mview'
html_page_title     = 'Reads_linking_16S_and_contig'
mafft_exe           = 'mafft'

mview_linkage(combined_ref_reads, html_page_title, mafft_exe, mview_exe)
